from pytvc.physics import *
import numpy as np
import unittest

DEG_TO_RAD = np.pi / 180

class test_vector_math(unittest.TestCase):

    def test_eq(self):
        self.assertEqual(Vec3(0.0, 0.0, 0.0), Vec3(0.0, 0.0, 0.0))

    def test_add(self):
        self.assertEqual(Vec3(0.0, 0.0, 0.0) +
                         Vec3(1.0, 1.0, 1.0), Vec3(1.0, 1.0, 1.0))

    def test_sub(self):
        self.assertEqual(Vec3(0.0, 0.0, 0.0) - Vec3(1.0,
                         1.0, 1.0), Vec3(-1.0, -1.0, -1.0))

    def test_mul(self):
        self.assertEqual(Vec3(1.0, 1.0, 1.0) *
                         Vec3(2.0, 2.0, 2.0), Vec3(2.0, 2.0, 2.0))

    def test_mul_scalar(self):
        self.assertEqual(Vec3(1.0, 1.0, 1.0) * 2.0, Vec3(2.0, 2.0, 2.0))

    def test_div(self):
        self.assertEqual(Vec3(1.0, 1.0, 1.0) /
                         Vec3(2.0, 2.0, 2.0), Vec3(0.5, 0.5, 0.5))

    def test_div_scalar(self):
        self.assertEqual(Vec3(1.0, 1.0, 1.0) / 2.0, Vec3(0.5, 0.5, 0.5))

    def test_norm(self):
        self.assertEqual(Vec3(5.0, 5.0, 5.0).normalize(), Vec3(
            0.5773502691896257, 0.5773502691896257, 0.5773502691896257))

    def test_len(self):
        self.assertEqual(Vec3(5.0, 5.0, 5.0).length(), 8.660254037844387)

    def test_abs(self):
        self.assertEqual(Vec3(1.0, 1.0, 1.0),
                         abs(Vec3(-1.0, -1.0, -1.0)))

    def test_round(self):
        self.assertEqual(Vec3(1.1, 1.1, 1.1),
                         round(Vec3(1.11, 1.11, 1.11), 1))

    def test_str(self):
        self.assertEqual("1.0, 1.0, 1.0", str(Vec3(1.0, 1.0, 1.0)))


class test_Quat_math(unittest.TestCase):

    def test_eq(self):
        self.assertEqual(Quat(1.0, 0.0, 0.0, 0.0),
                         Quat(1.0, 0.0, 0.0, 0.0))

    def test_mul(self):
        self.assertEqual(Quat(1.0, 0.5, 0.5, 0.5) * Quat(1.0,
                         0.5, 0.5, 0.5), Quat(0.25, 1.0, 1.0, 1.0))

    def test_conj(self):
        self.assertEqual(Quat(1.0, 0.5, 0.5, 0.5).conjugate(),
                         Quat(1.0, -0.5, -0.5, -0.5))

    def test_norm(self):
        q = Quat(2.0, 2.0, 2.0, 2.0)
        self.assertAlmostEqual(q.length(), 4.0, 4)

    def test_rotate(self):

        q: Quat = Quat().from_euler(Vec3(0, 90 * DEG_TO_RAD, 0))
        print(q)
        v: Vec3 = Vec3(1, 0, 0)

        vt = q.rotate(v)
        print(vt)

        self.assertAlmostEqual(vt.x, 0.0, 4)
        self.assertAlmostEqual(vt.y, 0.0, 4)
        self.assertAlmostEqual(vt.z, -1.0, 4)

    def test_euler_to_Quat(self):

        e = Vec3(45 * DEG_TO_RAD, 45 * DEG_TO_RAD, 45 * DEG_TO_RAD)
        q = Quat().from_euler(e)

        qt = Quat(0.8446231020115715, 0.19134170284356303,
                        0.4619399539487806, 0.19134170284356303)

        self.assertAlmostEqual(q.w, qt.w, 4)
        self.assertAlmostEqual(q.x, qt.x, 4)
        self.assertAlmostEqual(q.y, qt.y, 4)
        self.assertAlmostEqual(q.z, qt.z, 4)

    def test_Quat_to_euler(self):

        q = Quat(0.8446231020115715, 0.19134170284356303,
                       0.4619399539487806, 0.19134170284356303)

        e = q.to_euler()
        et = Vec3(45 * DEG_TO_RAD, 45 * DEG_TO_RAD, 45 * DEG_TO_RAD)
        # print(q)
        # print(e * RAD_TO_DEG)

        self.assertAlmostEqual(e.x, et.x, 4)
        self.assertAlmostEqual(e.y, et.y, 4)
        self.assertAlmostEqual(e.z, et.z, 4)


class test_body_math(unittest.TestCase):

    def test_rotation(self):

        body: physics_body = physics_body()

        for i in range(999):
            body.apply_torque(Vec3(0.0, 0.0, 1))
            body.update(0.001)
            body.clear()
        
        self.assertAlmostEqual(body.rotation_euler.z, 0.5, 1)

        for i in range(999):
            body.apply_torque(Vec3(0.0, 0.0, -1.0))
            body.update(0.001)
            body.clear()

        self.assertAlmostEqual(body.rotation_euler.z, 1.0, 1)

    def test_aero_forces(self):

        body: physics_body = physics_body(position=Vec3(1000.0, 0.0, 0.0), velocity=Vec3(-1.0, 0.0, 0.0), drag_coefficient_forewards=1.0, drag_coefficient_sideways=10.0, ref_area=1.0, use_aero=True)

        body.update(0.01)

        self.assertAlmostEqual(body.drag_force, Vec3(0.6125, 0.0, 0.0), 4)

        body.velocity = Vec3(-100.0, 0.0, 0.0)

        body.update(0.01)
        if isinstance(body.drag_force, Vec3):
            self.assertAlmostEqual(body.drag_force, Vec3(6125.0, 0.0, 0.0), 4)
        
if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from dataclasses import dataclass
import numpy as np
from pytvc.data import rocket_motor


def clamp(x, min_val, max_val):
    return max(min(x, max_val), min_val)


@dataclass
class Vec3:

    x: float = 0.0
    y: float = 0.0
    z: float = 0.0

    def __init__(self, x: float = 0.0, y: float = 0.0, z: float = 0.0) -> None:
        """__init__ initializes a Vec3 object.

        Args:
            x (float, optional): the x component of the vector. Defaults to 0.0.
            y (float, optional): the y component of the vector. Defaults to 0.0.
            z (float, optional): the z component of the vector. Defaults to 0.0.
        """
        self.x = x
        self.y = y
        self.z = z

    def __add__(self, other: Vec3) -> Vec3:
        """__add__ adds 2 vectors and returns the sum

        Args:
            other (Vec3): right hand vector to add

        Raises:
            NotImplementedError: raises notImplemented if the right hand side is not a vector object

        Returns:
            Vec3: a vector sum of the left hand and right hand vector
        """
        if isinstance(other, Vec3):
            return Vec3(self.x + other.x, self.y + other.y, self.z + other.z)
        else:
            raise NotImplementedError

    def __sub__(self, other: Vec3) -> Vec3:
        """__sub__ subtracts 2 vectors and returns the difference

        Args:
            other (Vec3): right hand vector to subtract

        Raises:
            NotImplementedError: raises notImplemented if the right hand side is not a vector object

        Returns:
            Vec3: a vector difference of the left hand and right hand vector
        """
        if isinstance(other, Vec3):
            return Vec3(self.x - other.x, self.y - other.y, self.z - other.z)
        else:
            raise NotImplementedError

    def __mul__(self, other: Vec3) -> Vec3:
        """__mul__ multiplies 2 vectors and returns the product

        Args:
            other (Vec3): right hand vector to multiply

        Raises:
            NotImplementedError: raises notImplemented if the right hand side is not a vector object

        Returns:
            Vec3: a vector product of the left hand and right hand vector
        """
        if isinstance(other, Vec3):
            return Vec3(self.x * other.x, self.y * other.y, self.z * other.z)
        elif isinstance(other, int) or isinstance(other, float):
            return Vec3(self.x * other, self.y * other, self.z * other)
        else:
            raise NotImplementedError

    def __truediv__(self, other) -> Vec3:
        """__truediv__ divides 2 vectors and returns the quotient

        Args:
            other: right hand vector to divide or scalar to divide by

        Raises:
            NotImplementedError: raises notImplemented if the right hand side is not a vector object or a number

        Returns:
            Vec3: a vector quotient of the left hand and right hand vector
        """
        if isinstance(other, Vec3):
            return Vec3(self.x / other.x, self.y / other.y, self.z / other.z)
        elif isinstance(other, int) or isinstance(other, float):
            return Vec3(self.x / other, self.y / other, self.z / other)
        else:
            raise NotImplementedError

    def __repr__(self) -> str:
        """__repr__ returns a string representation of the vector

        Returns:
            str: a string representation of the vector
        """
        return f"Vec3({self.x}, {self.y}, {self.z})"

    def __str__(self) -> str:
        """__str__ returns a string representation of the vector

        Returns:
            str: a string representation of the vector
        """
        return f"{self.x}, {self.y}, {self.z}"

    def __neg__(self) -> Vec3:
        """__neg__ returns the negative of the vector

        Returns:
            Vec3: the negative of the vector
        """
        return Vec3(-self.x, -self.y, -self.z)

    def __eq__(self, other: Vec3) -> bool:
        """__eq__ returns true if the vectors are equal

        Args:
            other (Vec3): right hand vector to compare

        Raises:
            NotImplementedError: raises notImplemented if the right hand side is not a vector object

        Returns:
            bool: true if the vectors are equal
        """
        if isinstance(other, Vec3):
            return self.x == other.x and self.y == other.y and self.z == other.z
        elif isinstance(other, int) or isinstance(other, float):
            return self.length() == other
        else:
            raise NotImplementedError

    def __ne__(self, other: Vec3) -> bool:
        """__ne__ returns true if the vectors are not equal

        Args:
            other (Vec3): right hand vector to compare

        Raises:
            NotImplementedError: raises notImplemented if the right hand side is not a vector object

        Returns:
            bool: true if the vectors are not equal
        """
        if isinstance(other, Vec3):
            return self.x != other.x or self.y != other.y or self.z != other.z
        else:
            print(other.__class__)
            raise NotImplementedError

    def __round__(self, ndigits: int = 0) -> Vec3:
        """__round__ returns the rounded vector

        Args:
            ndigits (int): number of digits to round to

        Returns:
            Vec3: the rounded vector
        """
        return Vec3(round(self.x, ndigits), round(self.y, ndigits), round(self.z, ndigits))

    def __abs__(self) -> Vec3:
        """__abs__ returns the absolute value of the vector

        Returns:
            Vec3: the absolute value of the vector
        """
        return Vec3(abs(self.x), abs(self.y), abs(self.z))

    def cross(self, other: Vec3) -> Vec3:
        """cross returns the cross product of 2 vectors

        Args:
            other (Vec3): right hand vector to cross

        Raises:
            NotImplementedError: raises notImplemented if the right hand side is not a vector object

        Returns:
            Vec3: a vector cross product of the left hand and right hand vector
        """
        if isinstance(other, Vec3):
            return Vec3(self.y * other.z - self.z * other.y, self.z * other.x - self.x * other.z, self.x * other.y - self.y * other.x)
        else:
            raise NotImplementedError

    def dot(self, other: Vec3) -> float:
        """dot returns the dot product of 2 vectors

        Args:
            other (Vec3): right hand vector to dot

        Raises:
            NotImplementedError: raises notImplemented if the right hand side is not a vector object

        Returns:
            float: a vector dot product of the left hand and right hand vector
        """
        if isinstance(other, Vec3):
            return self.x * other.x + self.y * other.y + self.z * other.z
        else:
            raise NotImplementedError

    def length(self) -> float:
        """length returns the length of the vector

        Returns:
            float: the length of the vector
        """
        return np.sqrt(self.x**2 + self.y**2 + self.z**2)

    def normalize(self) -> Vec3:
        """normalize returns the normalized vector

        Returns:
            Vec3: the normalized vector
        """
        return self / self.length()

    def angle_between_vectors(self, vector: Vec3) -> float:
        """Calculate the angle between two vectors."""
        if isinstance(vector, Vec3):
            return np.arccos(self.dot(vector) / (self.length() * vector.length()))
        else:
            return None


@dataclass
class Quat:

    w: float = 1.0
    x: float = 0.0
    y: float = 0.0
    z: float = 0.0

    def __init__(self, w: float = 1.0, x: float = 0.0, y: float = 0.0, z: float = 0.0):
        """__init__ initializes a Quat object

        Args:
            w (float, optional): the real component of the Quat. Defaults to 1.0.
            x (float, optional): the x component of the Quat. Defaults to 0.0.
            y (float, optional): the y component of the Quat. Defaults to 0.0.
            z (float, optional): the z component of the Quat. Defaults to 0.0.
        """
        self.w = w
        self.x = x
        self.y = y
        self.z = z

    def __mul__(self, other: Quat) -> Quat:
        """__mul__ multiplies 2 Quats and returns the product

        Args:
            other (Quat): right hand Quat to multiply

        Raises:
            NotImplementedError: raises notImplemented if the right hand side is not a Quat object

        Returns:
            Quat: a Quat product of the left hand and right hand Quat
        """
        if isinstance(other, Quat):
            return Quat(
                self.w * other.w - self.x * other.x - self.y * other.y - self.z * other.z,
                self.w * other.x + self.x * other.w + self.y * other.z - self.z * other.y,
                self.w * other.y - self.x * other.z + self.y * other.w + self.z * other.x,
                self.w * other.z + self.x * other.y - self.y * other.x + self.z * other.w
            )
        else:
            raise NotImplementedError

    def __truediv__(self, scalar: float) -> Quat:
        """__truediv__ _summary_

        Args:
            scalar (float): _description_

        Returns:
            Quat: _description_
        """

        if isinstance(scalar, float) or isinstance(scalar, int):
            return Quat(self.w / scalar, self.x / scalar, self.y / scalar, self.z / scalar)

    def __add__(self, other: Quat) -> Quat:
        """__add__ adds 2 Quats and returns the sum

        Args:
            other (Quat): right hand Quat to add

        Raises:
            NotImplementedError: raises notImplemented if the right hand side is not a Quat object

        Returns:
            Quat: a Quat sum of the left hand and right hand Quat
        """
        if isinstance(other, Quat):
            return Quat(self.w + other.w, self.x + other.x, self.y + other.y, self.z + other.z)
        else:
            raise NotImplementedError

    def __sub__(self, other: Quat) -> Quat:
        """__sub__ subtracts 2 Quats and returns the difference

        Args:
            other (Quat): right hand Quat to subtract

        Raises:
            NotImplementedError: raises notImplemented if the right hand side is not a Quat object

        Returns:
            Quat: a Quat difference of the left hand and right hand Quat
        """
        if isinstance(other, Quat):
            return Quat(self.w - other.w, self.x - other.x, self.y - other.y, self.z - other.z)
        else:
            raise NotImplementedError

    def __repr__(self) -> str:
        """__repr__ returns the Quat as a string

        Returns:
            str: the Quat as a string
        """
        return f"Quat({self.w}, {self.x}, {self.y}, {self.z})"

    def __str__(self) -> str:
        """__str__ returns the Quat as a string

        Returns:
            str: the Quat as a string
        """
        return f"Quat({self.w}, {self.x}, {self.y}, {self.z})"

    def length(self) -> float:
        """length returns the length of the Quat

        Returns:
            float: the length of the Quat
        """
        return np.sqrt(self.w**2 + self.x**2 + self.y**2 + self.z**2)

    def normalize(self) -> Quat:
        """normalize returns the normalized Quat

        Returns:
            Quat: the normalized Quat
        """
        return self / self.length()

    def conjugate(self) -> Quat:
        """conjugate returns the conjugate of the Quat

        Returns:
            Quat: the conjugate of the Quat
        """
        return Quat(self.w, -self.x, -self.y, -self.z)

    def rotate(self, vector: Vec3) -> Vec3:
        """rotate returns the rotated vector

        Args:
            vector (Vec3): vector to rotate

        Raises:
            NotImplementedError: raises notImplemented if the right hand side is not a vector object

        Returns:
            Vec3: the rotated vector
        """
        if isinstance(vector, Vec3):
            tmp: Quat = Quat(0, vector.x, vector.y, vector.z)
            tmp = self * tmp * self.conjugate()
            return Vec3(tmp.x, tmp.y, tmp.z)
        else:
            raise NotImplementedError(f"cannot rotate type {vector.__class__}")

    def fractional(self, a) -> Quat:
        """Return the fractional of the Quat."""

        self.w = 1-a + a*self.w
        self.x *= a
        self.y *= a
        self.z *= a

        return self.norm()

    def rotation_between_vectors(self, vector: Vec3) -> Quat:
        """Return the rotation Quat between two vectors.

        parameters:

        vector : vector3
            vector to rotate
        """

        q = Quat(0.0, vector.x, vector.y, vector.z)

        q = self * q
        q.w = 1 - q.w

        return q.normalize()

    def from_axis_angle(self, axis: Vec3, angle: float) -> Quat:
        """Return the Quat from an axis and angle.

        parameters:

        axis : vector3
            axis of rotation

        angle : float
            angle of rotation
        """

        sa: float = np.sin(angle / 2)

        self.w = np.cos(angle / 2)
        self.x = axis.x * sa
        self.y = axis.y * sa
        self.z = axis.z * sa

        return self

    def from_euler(self, euler_angles: Vec3) -> Quat:
        """set the Quat from euler angles

        Args:
            euler_angles (Vec3): euler angles to set to

        Raises:
            NotImplementedError: raises notImplemented if the right hand side is not a vector object

        Returns:
            Quat: the Quat set from euler angles
        """

        cr = np.cos(euler_angles.x / 2.0)
        cp = np.cos(euler_angles.y / 2.0)
        cy = np.cos(euler_angles.z / 2.0)

        sr = np.sin(euler_angles.x / 2.0)
        sp = np.sin(euler_angles.y / 2.0)
        sy = np.sin(euler_angles.z / 2.0)

        self.w = cr * cp * cy + sr * sp * sy
        self.x = sr * cp * cy - cr * sp * sy
        self.y = cr * sp * cy + sr * cp * sy
        self.z = cr * cp * sy - sr * sp * cy

        return self

    def to_euler(self) -> Vec3:
        """Convert a Quat to euler angles."""
        x = np.arctan2(2.0 * (self.w * self.x + self.y * self.z),
                       1.0 - 2.0 * (self.x**2 + self.y**2))
        y = np.arcsin(2.0 * (self.w * self.y - self.z * self.x))
        z = np.arctan2(2.0 * (self.w * self.z + self.x * self.y),
                       1.0 - 2.0 * (self.y**2 + self.z**2))

        return Vec3(x, y, z)


class TVC:

    def __init__(self, name: str = "", speed: float = 0.0, offset: Vec3 = Vec3(), throttleSpeed: float = 0.0, maxPosition: Vec3 = Vec3(), minPosition: Vec3 = Vec3(), minThrottle: float = 0.9, actuator_precision: float = 0.0, linkage_ratio: float = 0.0, update_delay: float = 0.02, log_data: bool = True, noise: float = 0.0) -> None:
        """__init__ initializes the TVC object

        Args:
            speed (float, optional): speed of the servos. Defaults to 0.0.
            offset (Vec3, optional): offset of the TVC mount. Defaults to Vec3.
            throttleSpeed (float, optional): speed at which the throttling mechanism throttles. Defaults to 0.0.
            maxPosition (Vec3, optional): maximum positive position of the mount. Defaults to Vec3.
            minPosition (Vec3, optional): maximum negative position of the mount. Defaults to Vec3.
            minThrottle (float, optional): minimum throttle value. Defaults to 0.9.
            actuator_precision (float, optional): precision of the servos, if set to zero it will be infinite. Defaults to 0.0.
            linkage_ratio (float, optional): ratio between servo movement and TVC movement, if set to zero will skip calculation. Defaults to 0.0.
        """

        self.name: str = name

        self._rotation: Quat = Quat()
        self._rotationEulers: Vec3 = Vec3()
        self._throttle: float = 1.0

        self.offset: Vec3 = offset
        self.maxPosition: Vec3 = maxPosition
        self.minPosition: Vec3 = minPosition
        self.targetEulers: Vec3 = Vec3()
        self.speed: float = speed
        self.offset: Vec3 = Vec3()
        self.targetThrottle: float = 1.0
        self.throttleSpeed: float = throttleSpeed
        self.minThrottle: float = minThrottle
        self.noise: float = noise

        self.log_data: bool = log_data
        self.update_delay = update_delay
        self.last_update: float = 0.0

        self.target_throttle: float = 1.0

        self._motors = {}
        self._update_func = None

    def update_func(self, func):
        """update_func sets the update function for the TVC

        the update function must return either a Vec3 class containing the TVC control output in radians, or a Quat class with a control output rotation

        Args:
            func (function) -> Vec3 or Quat: function to set as update function
        """
        self._update_func = func
        return func

    def add_motor(self, motor, name: str = "") -> None:
        """addMotor adds a motor to the TVC mount

        Args:
            motor (str or rocket_motor): _description_
            name (str, optional): name of the rocket motor, if no name is given the name defaults to motor + the motor index. Defaults to "".

        Raises:
            ValueError: if the motor is already added
            TypeError: the motor is not a rocket_motor or string
        """
        if isinstance(motor, rocket_motor):
            if name in self._motors:
                raise ValueError("A motor with the name " +
                                 name + " already exists")
            else:
                self._motors[name] = motor
        elif isinstance(motor, str):
            if name in self._motors:
                raise ValueError("A motor with the name " +
                                 name + " already exists")
            else:
                self._motors[name] = rocket_motor(motor, self.time_step)
        else:
            raise TypeError("motor must be a rocket_motor or a string")

    def throttle(self, target_throttle) -> None:
        self.target_throttle = target_throttle

    def set_position(self, target_position) -> None:
        if isinstance(target_position, Quat):
            self.targetEulers = target_position.to_euler()
        elif isinstance(target_position, Vec3):
            self.targetEulers = target_position
        else:
            raise NotImplemented

    def update(self, dt: float):
        """update updates the TVC mount and throttling mechanism

        Args:
            dt (float): time step
        """

        if self._update_func() is not None:
            self.set_position(self._update_func())

        error: Vec3 = self.targetEulers - self._rotationEulers
        error = error * self.speed * dt
        self._rotationEulers += error

        self._rotationEulers.y = clamp(
            self._rotationEulers.y, self.minPosition.y, self.maxPosition.y)
        self._rotationEulers.z = clamp(
            self._rotationEulers.z, self.minPosition.z, self.maxPosition.z)

        self._rotation = Quat().from_euler(self._rotationEulers + self.offset + Vec3(0.0,
                                                                                     np.random.normal(0.0, 1, 1)[0] * self.noise, np.random.normal(0.0, 1, 1)[0] * self.noise))
        self._rotationEulers = self._rotation.to_euler()
        self._rotationEulers.x = 0.0

        throttle_error = self.target_throttle - self._throttle
        throttle_error = throttle_error * self.throttleSpeed * dt
        self._throttle += throttle_error

        if self._throttle < self.minThrottle:
            self._throttle = self.minThrottle
        if self._throttle > 1.0:
            self._throttle = 1.0

    def get_values(self, time: float) -> tuple[float, float]:
        """get values returns the a tuple of the current mass and thrust vector from the TVC mount

        Args:
            time (float): the current time for the motors

        Returns:
            tuple[float, float]: a tuple of the current mass and thrust vector from the TVC mount
        """
        for motor in self._motors.values():
            tmp_t, tmp_m = motor.update(time)

        thrust: Vec3 = self._rotation.rotate(Vec3(tmp_t, 0.0, 0.0))

        return tmp_m, thrust


class parachute:

    def __init__(self, diameter: float = 1.0, cord_len: float = 1.0, drag_coeff: float = 1.75, name: str = "") -> None:
        """initializes the parachute

        Args:
            diameter (float, optional): diameter of the parachute in meters. Defaults to 1.0.
            cord_len (float, optional): length of the shock cord connecting to the rocket ( not implemented ). Defaults to 1.0.
            drag_coeff (float, optional): drag coefficient of the parachute. Defaults to 1.75.

        Returns:
            _type_: _description_
        """
        self.drag_area: float = np.pi * (diameter*diameter) / 4.0
        self.cord_len: float = cord_len
        # average value for a parachute from nasa
        self.drag_coefficient: float = drag_coeff

        self.name = name

        self._deploy_func = None

        return None

    def deploy_func(self, func):
        """sets the function to call when the parachute is deployed

        Args:
            func (function(position: Vec3, velocity: Vec3)): function to call when the parachute is deployed
        """
        self._deploy_func = func
        return func

    def _check(self, position: Vec3, velocity: Vec3) -> bool:
        """checks if the parachute is deployed

        Args:
            position (Vec3): the position of the parachute
            velocity (Vec3): the velocity of the parachute

        Returns:
            bool: True if the parachute is deployed, False otherwise
        """
        result = self._deploy_func(position, velocity)
        if isinstance(result, bool):
            return result
        else:
            raise TypeError("deploy_func must return a bool")

    def calculate_forces(self, mass: float, velocity: Vec3, air_density: float = 1.225) -> Vec3:

        # untested, probably broken

        force_g = Vec3(-mass * 9.806, 0.0, 0.0)
        force_d = -velocity.normalize() * (self.drag_coefficient/2.0 * air_density * self.drag_area * velocity.length()**2)

        return force_d

        # net_force_dir: Vec3 = (force_d) + force_g
        # net_force_dir = net_force_dir.normalize()

        # force_dir_q = Quat(1.0, 0.0, 0.0, 0.0).rotation_between_vectors(net_force_dir)

        # return force_dir_q.rotate(force_d)


class physics_body:

    """physics_body"""

    def __init__(self, position: Vec3 = Vec3(), velocity: Vec3 = Vec3(), rotation: Quat = Quat(), rotational_velocity: Vec3 = Vec3(),
                 mass: float = 1.0, moment_of_inertia: Vec3 = Vec3(1.0, 1.0, 1.0), ref_area: float = 1.0, drag_coefficient_forewards: float = 0.0,
                 drag_coefficient_sideways: float = 0.0, wind_speed: Vec3 = Vec3(), cp_location: Vec3 = Vec3(), friction_coeff: float = 0.0, use_aero: bool = False):
        """initializes the physics_body object

        Args:
            position (Vec3, optional): position of the body. Defaults to Vec3.
            velocity (Vec3, optional): velocity of the body. Defaults to Vec3.
            rotation (Quat, optional): rotation of the body. Defaults to Quat.
            rotational_velocity (Vec3, optional): rotational velocity of the body. Defaults to Vec3.
            mass (float, optional): mass of the body. Defaults to 1.0.
            moment_of_inertia (Vec3, optional): inertia of the body. Defaults to Vec3.
            ref_area (float, optional): reference area of the body. Defaults to 1.0.
            drag_coefficient_forewards (float, optional): drag coefficient of the body in the forewards direction. Defaults to 0.0.
            drag_coefficient_sideways (float, optional): drag coefficient of the body in the sideways direction. Defaults to 0.0.
            wind_speed (Vec3, optional): speed of wind relative to the body. Defaults to Vec3.
            cp_location (Vec3, optional): center of pressure location of the body. Defaults to Vec3.
            friction_coeff (float, optional): friction coefficient of the body. Defaults to 0.0.
            use_aero (bool, optional): whether or not to use aerodynamic forces. Defaults to False.
        """

        self.position: Vec3 = position
        self.velocity: Vec3 = velocity

        self.acceleration: Vec3 = Vec3(0.0, 0.0, 0.0)
        self.acceleration_local: Vec3 = Vec3(0.0, 0.0, 0.0)

        self.rotation: Quat = rotation
        self.rotation_euler: Vec3 = Vec3(0.0, 0.0, 0.0)

        self.rotational_velocity: Vec3 = rotational_velocity
        self.rotational_acceleration: Vec3 = Vec3(0.0, 0.0, 0.0)

        self.mass: float = mass
        self.moment_of_inertia: Vec3 = moment_of_inertia

        self.aoa: float = 0.0
        self.ref_area: float = ref_area

        self.drag_coefficient_forwards: float = drag_coefficient_forewards
        self.drag_coefficient_sideways: float = drag_coefficient_sideways

        self.wind_speed: Vec3 = wind_speed

        self.cp_location: Vec3 = cp_location

        self.friction_coeff: float = friction_coeff
        self.air_density: float = 1.225

        self.drag_force: Vec3 = Vec3()

        self.use_aero: bool = use_aero

    def apply_torque(self, torque: Vec3) -> None:
        """apply a torque on the body in the global frame

        Args:
            torque (Vec3): _description_
        """
        self.rotational_acceleration += (torque / self.moment_of_inertia)

    def apply_local_torque(self, torque: Vec3) -> None:
        """apply a torque in the local frame

        Args:
            torque (Vec3): the torque to apply
        """
        self.apply_torque(self.rotation.rotate(torque))

    def apply_point_torque(self, force: Vec3, point: Vec3) -> None:
        """apply a point torque in the global frame

        Args:
            torque (Vec3): the force to apply
            point (Vec3): distance of the force from the center of mass
        """
        tmp = point.cross(force)
        self.apply_torque(tmp)

    def apply_local_point_torque(self, force: Vec3, point: Vec3) -> None:
        """apply a point torque in the local frame

        Args:
            force (Vec3): the force to apply 
            point (Vec3): the distance of the force from the center of mass
        """
        self.apply_point_torque(self.rotation.rotate(force), self.rotation.rotate(point))

    def apply_force(self, force: Vec3) -> None:
        """apply a force on the body in the global frame

        Args:
            force (Vec3): the force to apply
        """
        accel: Vec3 = force / self.mass
        self.acceleration += accel
    
    def apply_local_force(self, force: Vec3) -> None:
        """apply a force in the local frame

        Args:
            force (Vec3): the force to apply
        """
        self.apply_force(self.rotation.rotate(force))

    def apply_point_force(self, force: Vec3, point: Vec3) -> None:
        """apply a point force in the global frame, affects rotation and translation

        Args:
            force (Vec3): the force to apply
            point (Vec3): the distance of the force from the center of mass
        """
        self.apply_force(force)
        self.apply_point_torque(force, point)

    def apply_local_point_force(self, force: Vec3, point: Vec3) -> None:
        """apply a point force in the local frame, affects rotation and translation

        Args:
            force (Vec3): the force to apply
            point (Vec3): the distance of the force from the bodies center of mass
        """
        self.apply_point_force(self.rotation.rotate(force), self.rotation.rotate(point))

    def update(self, dt: float) -> None:
        """updates the physics body for the given change in time

        Args:
            dt (float): change in time from the previous update to "now"
        """

        # aero
        velocity_relative_wind: Vec3 = self.velocity - self.wind_speed
        if velocity_relative_wind.length() > 0.0:

            self.aoa = velocity_relative_wind.angle_between_vectors(
                self.rotation.rotate(Vec3(1.0, 0.0, 0.0)))

            if self.aoa <= 1e-5:
                self.aoa = 1e-5

            if (self.aoa > 1.5708):
                self.aoa = np.pi - self.aoa

            drag_coefficient = self.drag_coefficient_forwards + \
                ((-np.cos(self.aoa)/2.0 + 0.5) *
                 (self.drag_coefficient_sideways - self.drag_coefficient_forwards))

            self.drag_force = -velocity_relative_wind.normalize() * (drag_coefficient/2.0 *
                                                                     self.air_density * self.ref_area * (self.velocity.length()*self.velocity.length()))

            self.apply_point_force(self.drag_force, self.cp_location)
            self.apply_torque(self.rotational_velocity * -
                              self.friction_coeff * self.air_density)

        self.acceleration_local = self.rotation.conjugate().rotate(self.acceleration)

        self.acceleration.x -= 9.81

        self.position += self.velocity * dt + (self.acceleration*0.5) * (dt*dt)
        self.velocity += self.acceleration * dt

        self.rotational_velocity += self.rotational_acceleration * dt

        ang: float = self.rotational_velocity.length()
        if ang > 0.0:
            self.rotation *= Quat().from_axis_angle(self.rotational_velocity / ang, ang * dt)

        self.rotation_euler = self.rotation.to_euler()

        if self.position.x < 0.0:
            self.position.x = 0.0
            self.velocity = Vec3()
            # self.rotational_velocity = Vec3()

    def clear(self):
        """clears the physics body of all forces and torques"""
        self.acceleration = Vec3(0.0, 0.0, 0.0)
        self.rotational_acceleration = Vec3(0.0, 0.0, 0.0)



import pytvc.control
import pytvc.core
import pytvc.data
import pytvc.physics
from pytvc.physics import Vec3, Quat
import pathlib

cPath = f"{pathlib.Path(__file__).parent.resolve()}"

H13=f"{cPath}/motors/AeroTech_H13ST.rse"
F10=f"{cPath}/motors/Apogee_F10.rse"
C6=f"{cPath}/motors/Estes_C6.rse"
D12=f"{cPath}/motors/Estes_D12.rse"
E12=f"{cPath}/motors/Estes_E12.rse"
F15=f"{cPath}/motors/Estes_F15.rse"



from pytvc.physics import Vec3
import numpy as np


class PID:

    """PID controller class

    for more information on the PID controller, see:
    https://en.wikipedia.org/wiki/PID_controller"""

    def __init__(self, Kp: float = 0.0, Ki: float = 0.0, Kd: float = 0.0, setpoint: float = 0.0, i_max: float = 0.0) -> None:
        """__init__ initializes the PID controller

        Args:
            Kp (float, optional): the proportional gain of the controller. Defaults to 0.0.
            Ki (float, optional): the integral gain of the controller. Defaults to 0.0.
            Kd (float, optional): the derivative gain of the controller. Defaults to 0.0.
            setpoint (float, optional): the setpoint of the controller. Defaults to 0.0.
            i_max (float, optional): the maximum value for the integral component of the controller. Defaults to 0.0.
        """
        self.Kp = Kp
        self.Ki = Ki
        self.Kd = Kd
        self.i_max = i_max
        self.setpoint = setpoint
        self.i: float = 0.0
        self.last_error: float = 0.0

        self.output: float = 0.0

    def update(self, input: float, dt: float = 1.0, input_derivitive: float = 0.0) -> None:
        error = self.setpoint - input

        if self.i > self.i_max:
            self.i = self.i_max
        elif self.i < -self.i_max:
            self.i = -self.i_max

        d = (error - self.last_error) / dt
        self.last_error = error

        if input_derivitive != 0.0:
            d = input_derivitive

        self.output = (self.Kp * error) + (self.Ki * self.i) + (self.Kd * d)

    def reset(self) -> None:
        self.i = 0.0
        self.last_error = 0.0

    def setSetpoint(self, setpoint: float) -> None:
        """setSetpoint sets the setpoint of the PID controller

        Args:
            setpoint (float): setpoint of the PID controller
        """
        self.setpoint = setpoint

    def setKp(self, Kp: float) -> None:
        """setKp sets the proportional gain of the PID controller

        Args:
            Kp (float): proportional gain of the PID controller
        """
        self.Kp = Kp

    def setKi(self, Ki: float) -> None:
        """setKi sets the integral gain of the PID controller

        Args:
            Ki (float): integral gain of the PID controller
        """
        self.Ki = Ki

    def setKd(self, Kd: float) -> None:
        """setKd sets the derivative gain of the PID controller

        Args:
            Kd (float): derivative gain of the PID controller
        """
        self.Kd = Kd

    def setImax(self, i_max: float) -> None:
        """setImax set the maximum integral value for the PID controller

        Args:
            i_max (float): maximum value for the integral
        """
        self.i_max = i_max

    def getOutput(self) -> float:
        """getOutput returns the output of the PID controller

        Returns:
            float: output of the PID controller
        """
        return self.output


class torque_PID(PID):

    def __init__(self, Kp: float = 0.0, Ki: float = 0.0, Kd: float = 0.0, setpoint: float = 0.0, i_max: float = 0.0, inertia: float = 1.0, lever_arm: float = 1.0) -> None:
        """__init__ initializes the PID controller

        Args:
            Kp (float, optional): the proportional gain of the controller. Defaults to 0.0.
            Ki (float, optional): the integral gain of the controller. Defaults to 0.0.
            Kd (float, optional): the derivative gain of the controller. Defaults to 0.0.
            setpoint (float, optional): the setpoint of the controller. Defaults to 0.0.
            i_max (float, optional): the maximum value for the integral component of the controller. Defaults to 0.0.
            inertia (float, optional): the inertia to pass into the torque controller. Defaults to 1.0.
            lever_arm (float, optional): the lever arm to pass into the torque controller. Defaults to 1.0.
        """

        super().__init__(Kp, Ki, Kd, setpoint, i_max)
        self.inertia = inertia
        self.lever_arm = lever_arm

    def update(self, input: float, dt: float = 1.0, force: float = 1.0, input_derivitive: float = 0.0) -> None:
        """updates the PID controller

        Args:
            input (float): the input to the controller
            dt (float, optional): the time between this call and the last call of the PID controller. Defaults to 1.0.
            force (float, optional): the force that the controller is changing. Defaults to 1.0.
            input_derivitive (float, optional): an override of the built in derivitive. Defaults to 0.0.
        """
        if input_derivitive == 0.0:
            super().update(input, dt)
        else:
            super().update(input, dt, input_derivitive)

        if force != 0.0:
            calcval = ((super().getOutput() * self.inertia) / force) / self.lever_arm
            print(calcval)
            if abs(calcval) > 1.0:
                self.output = 0.0
            else:
                self.output = np.arcsin(calcval)
        else:
            self.output = 0.0



from pytvc.physics import *
from pytvc.data import data_logger, plotter, progress_bar
import numpy as np
import os

RAD2DEG = 180.0/np.pi

class rocket:

    def __init__(self, dry_mass: float = 1.0, time_step: float = 0.001, name: str = "", log_values: bool = False, log_frequency: int = 50, print_info: bool = False, do_aero: bool = True, drag_coeff_forewards: float = 0.0, drag_coeff_sideways: float = 0.0, reference_area: float = 0.0, cp_location: Vec3 = Vec3(), moment_of_inertia: Vec3 = Vec3(1.0, 1.0, 1.0), friction_coeff: float = 0.0) -> None:
        """__init__ initializes the rocket

        Args:
            dry_mass (float, optional): the mass of the vehicle with no motors. Defaults to 1.0.
            time_step (float, optional): time between updates. Defaults to 0.001.
            name (str, optional): name of the rocket. Defaults to "".
            log_values (bool, optional): if true the rocket will log values and export it to a CSV. Defaults to False.
            log_frequency (int, optional): frequency at which to log data. Defaults to 50.
            print_info (bool, optional): if true the rocket will print debug information while the simulation is initializing and running. Defaults to False.
            do_aero (bool, optional): enable aerodynamics simulation. Defaults to True.
            drag_coeff_forewards (float, optional): the drag coefficient when the rocket is pointed into the airstream. Defaults to 0.0.
            drag_coeff_sideways (float, optional): the drag coefficient when the rocket is at an angle of attack of 90 degrees. Defaults to 0.0.
            refence_area (float, optional): aerodynamic reference area of the rocket. Defaults to 0.0.
            cp_location (Vec3, optional): location of the center of pressure on the rocket. Defaults to Vec3().
            moment_of_inertia (Vec3, optional): moment of inertia of the rocekt. Defaults to Vec3(1.0, 1.0, 1.0).
        """

        # timing and variables
        self.print_info = print_info
        self.log_values = log_values
        self.log_delay: float = 1 / log_frequency
        self.last_log_time: float = 0
        self.name = name
        self.time = 0.0
        self.time_step = time_step
        self.apogee = 0.0

        # components
        self._n_tvc_mounts: int = 0
        self._n_parachutes: int = 0
        self._tvc_mounts = {}
        self._parachutes = {}

        self.datalogger: data_logger = data_logger()

        # physical properties
        self.dry_mass = dry_mass
        self.engine_mass = 0.0
        self.motor_thrust = 0.0

        self.body: physics_body = physics_body(
            use_aero=do_aero,
            drag_coefficient_forewards=drag_coeff_forewards,
            drag_coefficient_sideways=drag_coeff_sideways,
            moment_of_inertia=moment_of_inertia,
            mass=dry_mass,
            ref_area=reference_area,
            cp_location=cp_location,
            friction_coeff=friction_coeff
        )

        # user defined data points
        self._data_points = []

        # function registry for update and setup
        self._function_registry = {}

    def setup(self, func):
        self._function_registry["setup"] = func
        return func

    def update(self, func):
        self._function_registry["update"] = func
        return func

    def add_data_point(self, name: str, func: callable):
        """add_data_point add a data point to the rocket's datalogger

        Args:
            name (str): name of the data
            func (callable): function to call to get the data
        """
        self._data_points.append({"name": name, "func": func})

    def add_tvc_mount(self, tvc_mount: TVC, position: Vec3) -> None:
        """adds a TVC mount to the rocket

        Args:
            tvc_mount (TVC): TVC mount to add
            position (Vec3): position from the center of mass
            name (str): name of the mount

        Raises:
            Exception: _description_
            Exception: _description_
        """

        if isinstance(tvc_mount, TVC):
            self._n_tvc_mounts += 1
            if tvc_mount.name == "":
                tvc_mount.name = f"TVC{self._n_tvc_mounts}"

            if tvc_mount.name in self._tvc_mounts:
                raise Exception(
                    f"TVC mount with name {tvc_mount.name} already exists")
            else:
                self._tvc_mounts[tvc_mount.name] = {
                    "mount": tvc_mount, "position": position}
        else:
            raise Exception("TVC mount must be of type TVC")

    def add_parachute(self, chute: parachute, position: Vec3):

        if isinstance(chute, parachute):
            self._n_parachutes += 1
            if chute.name == "":
                chute.name = f"parachute_{self._n_parachutes}"
            self._parachutes[chute.name] = {
                "parachute": chute, "position": position}
        else:
            raise Exception("Parachute must be of type parachute")

    def _initialize(self) -> None:
        if self.log_values:
            if self.print_info:
                print(f"    {self.name}: initializing data logger")
            self.datalogger.add_datapoint("time")
            self.datalogger.add_datapoint("position_x")
            self.datalogger.add_datapoint("position_y")
            self.datalogger.add_datapoint("position_z")
            self.datalogger.add_datapoint("velocity_x")
            self.datalogger.add_datapoint("velocity_y")
            self.datalogger.add_datapoint("velocity_z")
            self.datalogger.add_datapoint("acceleration_x_local")
            self.datalogger.add_datapoint("acceleration_y_local")
            self.datalogger.add_datapoint("acceleration_z_local")
            self.datalogger.add_datapoint("acceleration_x_world")
            self.datalogger.add_datapoint("acceleration_y_world")
            self.datalogger.add_datapoint("acceleration_z_world")
            self.datalogger.add_datapoint("angular_velocity_x")
            self.datalogger.add_datapoint("angular_velocity_y")
            self.datalogger.add_datapoint("angular_velocity_z")
            self.datalogger.add_datapoint("rotation_x")
            self.datalogger.add_datapoint("rotation_y")
            self.datalogger.add_datapoint("rotation_z")
            self.datalogger.add_datapoint("motor_thrust")
            self.datalogger.add_datapoint("engine_mass")
            self.datalogger.add_datapoint("dry_mass")
            self.datalogger.add_datapoint("mass")

            for tvc in self._tvc_mounts:
                if self.print_info:
                    print(f"    {self.name}: adding TVC mount {tvc}")
                t = self._tvc_mounts[tvc]["mount"]
                if isinstance(t, TVC):
                    if t.log_data:
                        if self.print_info:
                            print(
                                f"    {self.name}: initializing TVC mount {t.name} data logger")
                        self.datalogger.add_datapoint(f"{t.name}_position_y")
                        self.datalogger.add_datapoint(f"{t.name}_position_z")
                        self.datalogger.add_datapoint(f"{t.name}_throttle")
                        self.datalogger.add_datapoint(f"{t.name}_setpoint_y")
                        self.datalogger.add_datapoint(f"{t.name}_setpoint_z")
                        self.datalogger.add_datapoint(f"{t.name}_error_y")
                        self.datalogger.add_datapoint(f"{t.name}_error_z")

            if len(self._data_points) != 0:
                for dp in self._data_points:
                    if self.print_info:
                        print(f"    {self.name}: adding data point {dp['name']}")
                    self.datalogger.add_datapoint(dp["name"])
            
            self.datalogger.fileName = f"{self.name}_log.csv"
            self.datalogger.initialize_csv(True, True)
            
            if self.print_info:
                print(f"    {self.name}: data logger initialized")

        if "setup" in self._function_registry:
            if self.print_info:
                print(f"    {self.name}: running user setup function")
            self._function_registry["setup"]()           

    def _update(self) -> None:
        """update the body's simulation by one time step"""

        self.time += self.time_step

        if self.body.position.x < 0.01 and self.body.velocity.x < 0.0:
            self._grounded = True

        tmp_m = 0.0
        for mount in self._tvc_mounts:
            self._tvc_mounts[mount]["mount"].update(self.time_step)
            tm, f = self._tvc_mounts[mount]["mount"].get_values(self.time)
            tmp_m += tm*0.001
            self.motor_thrust += f.length()
            self.body.apply_local_point_force(
                f, self._tvc_mounts[mount]["position"])

        if self.body.position.x > 0.01:
            for p in self._parachutes:
                if self._parachutes[p]["parachute"]._check(self.body.position, self.body.velocity):
                    self.body.apply_force(self._parachutes[p]["parachute"].calculate_forces(self.body.mass, self.body.velocity))# , self._parachutes[p]["position"])

        self.body.mass = self.dry_mass + tmp_m
        self.engine_mass = tmp_m

        self.body.update(self.time_step)

        if self.body.position.x > self.apogee:
            self.apogee = self.body.position.x

        if "update" in self._function_registry:
            self._function_registry["update"]()

        if self.log_values and self.time > self.last_log_time + self.log_delay:
            self.datalogger.record_variable("time", self.time)
            self.datalogger.record_variable("position_x", self.body.position.x)
            self.datalogger.record_variable("position_y", self.body.position.y)
            self.datalogger.record_variable("position_z", self.body.position.z)
            self.datalogger.record_variable("velocity_x", self.body.velocity.x)
            self.datalogger.record_variable("velocity_y", self.body.velocity.y)
            self.datalogger.record_variable("velocity_z", self.body.velocity.z)
            self.datalogger.record_variable("acceleration_x_local", self.body.acceleration_local.x)
            self.datalogger.record_variable("acceleration_y_local", self.body.acceleration_local.y)
            self.datalogger.record_variable("acceleration_z_local", self.body.acceleration_local.z)
            self.datalogger.record_variable("acceleration_x_world", self.body.acceleration.x)
            self.datalogger.record_variable("acceleration_y_world", self.body.acceleration.y)
            self.datalogger.record_variable("acceleration_z_world", self.body.acceleration.z)
            self.datalogger.record_variable("rotation_x", self.body.rotation_euler.x * RAD2DEG)
            self.datalogger.record_variable("rotation_y", self.body.rotation_euler.y * RAD2DEG)
            self.datalogger.record_variable("rotation_z", self.body.rotation_euler.z * RAD2DEG)
            self.datalogger.record_variable("angular_velocity_x", self.body.rotational_velocity.x)
            self.datalogger.record_variable("angular_velocity_y", self.body.rotational_velocity.y)
            self.datalogger.record_variable("angular_velocity_z", self.body.rotational_velocity.z)
            self.datalogger.record_variable("motor_thrust", self.motor_thrust)
            self.datalogger.record_variable("engine_mass", self.engine_mass)
            self.datalogger.record_variable("dry_mass", self.dry_mass)
            self.datalogger.record_variable("mass", self.body)

            for tvc in self._tvc_mounts:
                if self._tvc_mounts[tvc]["mount"].log_data:
                    t = self._tvc_mounts[tvc]["mount"]
                    self.datalogger.record_variable(f"{t.name}_position_y", t._rotationEulers.y * RAD2DEG)
                    self.datalogger.record_variable(f"{t.name}_position_z", t._rotationEulers.z * RAD2DEG)
                    self.datalogger.record_variable(f"{t.name}_throttle", t._throttle)
                    self.datalogger.record_variable(f"{t.name}_setpoint_y", t.targetEulers.y * RAD2DEG)
                    self.datalogger.record_variable(f"{t.name}_setpoint_z", t.targetEulers.z * RAD2DEG)
                    self.datalogger.record_variable(f"{t.name}_error_y", 0.0)
                    self.datalogger.record_variable(f"{t.name}_error_z", 0.0)
            
            if len(self._data_points) != 0:
                for dp in self._data_points:
                    self.datalogger.record_variable(dp["name"], dp["func"]())
            
            self.datalogger.save_data(True)
            self.last_log_time = self.time

        self.body.clear()


class simulation:
    """ simulation class"""

    def __init__(self, plot_data: bool = True, wind_speed: float = 0.5, time_step: float = 0.001, time_end: float = 60.0, rockets: dict = {}, print_times: bool = True, print_info: bool = True, wind: Vec3 = Vec3(), sim_wind: bool = False, random_wind: bool = False) -> None:
        self.time: float = 0.0
        self.time_step = time_step
        self.time_end = time_end
        self.nRockets = 0
        self._rockets = {}

        self.plot_data = plot_data
        self.print_times = print_times
        self.print_info = print_info

        self.use_wind = sim_wind

        if random_wind:
            self.wind: Vec3 = Vec3(0.0, np.random.normal(0.0, wind_speed, 1)[
                                   0], np.random.normal(0.0, wind_speed, 1)[0])
        else:
            self.wind: Vec3 = wind

        for r in rockets:
            if isinstance(r, rocket):
                if r.name == "":
                    r.name = "rocket" + str(self.nRockets)
                if sim_wind:
                    r.body.wind_speed = self.wind
                self._rockets[r.name] = r
                self.nRockets += 1
            else:
                raise TypeError("Rocket must be a rocket class")
        return None

    def add_rocket(self, r: rocket) -> None:
        if isinstance(r, rocket):
            if r.name == "":
                r.name = "rocket" + str(self.nRockets)
            self._rockets[r.name] = r
            self.nRockets += 1
        else:
            raise TypeError("Rocket must be a rocket class")

    def run(self) -> None:

        if self.print_info:
            print("Starting simulation...")

            print(f"""sim info:

rocket count: {self.nRockets}
wind speed: {round(self.wind, 2)}""")

        if self.nRockets == 0:
            raise Exception("No rockets to simulate!")

        for r in self._rockets:
            if self.print_info:
                print(f"Initializing {r}...")
                self._rockets[r].print_info = True
            self._rockets[r]._initialize()
            if self.print_info:
                print(f"{r} initialized")

        while self.time < self.time_end:

            self.time += self.time_step
            if self.time % 0.1 < self.time_step:
                print("Running simulation " +
                      progress_bar(self.time, self.time_end+self.time_step), end="\r")

            for r in self._rockets:
                self._rockets[r]._update()

        if self.plot_data:

            print("")

            for r in self._rockets:
                _rocket: rocket = self._rockets[r]
                if isinstance(_rocket, rocket):

                    pTmp: plotter = plotter()
                    pTmp.read_header(f"{_rocket.name}_log.csv")

                    pTmp.create_2d_graph(['time', 'position_x', 'position_y', 'position_z', 'velocity_x', 'velocity_y', 'velocity_z',
                                         'acceleration_x_world', 'acceleration_y_world', 'acceleration_z_world'], "x", "y", True, posArg=221)
                    pTmp.create_3d_graph(
                        ['position_x', 'position_y', 'position_z'], size=_rocket.apogee, posArg=122)
                    
                    for tvc in _rocket._tvc_mounts:
                        t = _rocket._tvc_mounts[tvc]["mount"]
                        if isinstance(t, TVC):
                            pTmp.create_2d_graph(['time', 'rotation_x', 'rotation_y', 'rotation_z',
                                                 f'{t.name}_position_y', f'{t.name}_position_z', f'{t.name}_setpoint_y', f'{t.name}_setpoint_z'], "x", "y", True, posArg=223)
                        else:
                            pTmp.create_2d_graph(
                                ['time', 'rotation_x', 'rotation_y', 'rotation_z'], "x", "y", True)
                    
                    # if len(_rocket._data_points) != 0:
                    #     l: list = ['time']
                    #     for dp in _rocket._data_points:
                    #         l.append(dp["name"])

                    #     print(l)

                    pTmp.show_all_graphs()
                    os.remove(f"{_rocket.name}_log.csv")

        return None



import setuptools

setuptools.setup(
    name="pytvc",
    version="0.1.2",
    description="Python simulation tool for TVC",
    author="ZegesMenden",
    packages=["pytvc"],
    scripts=["pytvc/control.py", "pytvc/core.py", "pytvc/data.py", "pytvc/physics.py"],
    include_package_data=True
)


import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d.axes3d as p3
import matplotlib.animation as animation
import xml.etree.ElementTree as ET
import numpy as np
import os
import csv

# from pytvc.physics import Vec3, Quat


def progress_bar(n, maxn) -> str:
    
    progress: str = ""
    end = int(n/maxn*50)
    
    for i in range(50):
        if i > end:
            progress += ' '
        else:
            if i+1 > end:
                progress += '>'
            else:
                progress += "="

    return (f"""[{progress}] Progress: {round((n/maxn)*100, 2)}%""")


class data_logger:
    def __init__(self):
        self.variables = 0
        self.variableDescriptions = []
        self.currentLog = {}
        self.loggedData = []
        self.initialized = False
        self.fileName = ""
        pass

    def add_datapoint(self, variableName):
        """Adds a data point to the logger object. Datapoints are added sequentially, so add your variables in the same sequence that you want them to show up in on the CSV"""
        if self.initialized == False:
            if str(variableName) in self.currentLog:
                raise IndexError("datapoiont already initialized")
            else:
                self.variables += 1
                self.variableDescriptions.append(variableName)
                self.currentLog[variableName] = None
        else:
            raise IndexError("file already initialized!")

    def record_variable(self, variableName, data):
        """records a variable to the current log, DOES NOT LOG AUTOMATICALLY"""
        if str(variableName) in self.currentLog:
            # if self.currentLog[str(variableName)] != None:
            #     raise Warning(f'data point {str(variableName)} is being overwritten!')
            self.currentLog[str(variableName)] = data
        else:
            raise IndexError("datapoint not initialized")

    def initialize_csv(self, makeFile, overWrite):
        """Initializes the CSV file and prepares it for writing."""
        self.initialized = True

        os.chdir(os.path.dirname(os.path.abspath(__file__)))

        if os.path.exists(str(self.fileName)):

            f = open(str(self.fileName), "r")

            # if the file is empty?
            if not f.read():
                f.close()

                f = open(str(self.fileName), "w")
                outString = ""
                for varName in self.variableDescriptions:
                    outString += varName
                    outString += ","

                f.write(outString[0:-1])

                f.write('\n')
                f.close()
            else:
                if overWrite == True:
                    f.close()

                    f = open(str(self.fileName), "w")
                    outString = ""
                    for varName in self.variableDescriptions:
                        outString += varName
                        outString += ","

                    f.write(outString[0:-1])

                    f.write('\n')
                if overWrite == False:
                    raise OSError("csv file is not empty!")

        else:
            if makeFile == True:
                f = open(str(self.fileName), "w")
                outString = ""
                for varName in self.variableDescriptions:
                    outString += varName
                    outString += ","

                f.write(outString[0:-1])

                f.write('\n')
                f.close()
            else:
                raise OSError("csv file not found!")

    def save_data(self, clearData):
        outString = ""
        for datapoint in self.currentLog:
            currentVar = self.currentLog[str(datapoint)]
            if currentVar == None:
                outString += "0"
            else:
                outString += str(currentVar)
            outString += ","
            if clearData == True:
                self.currentLog[str(datapoint)] = None
        f = open(str(self.fileName), "a")
        f.write(outString[0:-1] + "\n")

    def get_var(self, variableName):
        if str(variableName) in self.currentLog:
            return self.currentLog[str(variableName)]
        else:
            raise IndexError("datapoint not initialized")


class data_visualizer:
    def __init__(self):
        self.allDataDescriptions = []

    def graph_from_csv(self, datapoints, file_name):
        descriptionNum = 0
        pointsToLog = []

        for description in self.allDataDescriptions:
            for requestedDatapoint in datapoints:
                if str(description) == str(requestedDatapoint):
                    pointsToLog.append(descriptionNum)
            descriptionNum += 1

        with open(file_name, newline='\n') as pathFile:
            reader = csv.reader(pathFile, delimiter=',', quotechar='"')
            logList = []
            dataOut = []
            for index, row in enumerate(reader):
                for point in pointsToLog:
                    logList.append(row[point])
                if index == 0:
                    for x in logList:
                        dataOut.append([])

                if index > 0:
                    for idx, point in enumerate(dataOut):
                        point.append(float(row[pointsToLog[idx]]))
                logList = []

        return dataOut


class plotter:

    """Data visualization class"""

    def __init__(self) -> None:
        """initializes data visualizer"""

        self.data_descriptions: list = []
        self.header: str = ""
        self.viewer: data_visualizer = data_visualizer()

        self.n_plots: int = 0

    def read_header(self, file_name):
        self.file_name = file_name
        f = open(file_name, "r")
        self.header = f.readline()
        self.data_descriptions = self.header.split(',')
        self.viewer.allDataDescriptions = self.data_descriptions

    def create_2d_graph(self, graph_points: list, x_desc: str = "", y_desc: str = "", annotate: bool = False, posArg: int = 0):
        """Creates a 2-dimensional graph from the provided data descriptions.

        Params:

        graph_points - a list of all data point names to be graphed. First element of graph_points will be the x-axis datapoint.

        x_desc - label for the x axis of the graph.

        y_desc - label for the y axis of the graph.

        annotate - set true to show a color code for all lines on the graph."""

        if posArg != 0:
            if self.n_plots == 0:
                plt.figure(figsize=(16, 8))
                self.n_plots += 1
            plt.subplot(posArg)
        else:
            self.n_plots += 1
            plt.figure(self.n_plots)

        plot_points = self.viewer.graph_from_csv(graph_points, self.file_name)

        for index, dataPoint in enumerate(plot_points):
            if index > 0:
                if annotate:
                    plt.plot(plot_points[0], dataPoint,
                             label=graph_points[index])
                else:
                    plt.plot(plot_points[0], dataPoint)
        if annotate:
            plt.legend()
        plt.xlabel(x_desc)
        plt.ylabel(y_desc)

    def create_3d_graph(self, graph_points: list, size: float = 0.0, color: str = 'Blues', posArg: int = 0):
        """Creates a 3 dimensional graph from the provided data descriptions

        Params:

        graph_points - a list of all data point names to be graphed.

        size - the size of the graph in all axes"""

        if len(graph_points) != 3:
            raise IndexError("graph_points must have 3 data points!")

        plot_points = self.viewer.graph_from_csv(graph_points, self.file_name)

        if posArg != 0:
            if self.n_plots == 0:
                plt.figure(figsize=(16, 8))
                self.n_plots += 1
            plt.subplot(posArg, projection='3d', xlim=(-size, size),
                        ylim=(-size, size), zlim=(-0, size))
            plt.plot(plot_points[2], plot_points[1], plot_points[0])
        else:
            self.n_plots += 1
            plt.figure(self.n_plots)

            ax = plt.axes(projection='3d')

            if size != 0.0:

                ax.set_xlim3d(-size, size)
                ax.set_ylim3d(-size, size)
                ax.set_zlim3d(0, size)

            ax.scatter3D(plot_points[2], plot_points[1],
                         plot_points[0], c=plot_points[2], cmap=color)

    def create_3d_animation(self, graph_points: list, size: float, time: float = 5.0, color: str = 'Blues'):

        self.n_plots += 1
        fig = plt.figure(self.n_plots)

        ax = p3.Axes3D(fig)

        plot_position = self.viewer.graph_from_csv(
            graph_points, self.file_name)

        ax.set_xlim3d(-size, size)
        ax.set_ylim3d(-size, size)
        ax.set_zlim3d(0, size)

        def func(num, dataSet, line):
            # NOTE: there is no .set_data() for 3 dim data...
            xcur = dataSet[2][num] - 1
            numpog = 0
            for index, x in enumerate(dataSet[2]):
                if x >= xcur-0.1 and x <= xcur - 0.1:
                    numpog = index
            line.set_data(dataSet[0:2, num-20:num])
            line.set_3d_properties(dataSet[2, num-20:num])
            return line

        # # THE DATA POINTS
        # This would be the z-axis ('t' means time here)
        t = np.array(plot_position[1])
        x = np.array(plot_position[2])
        y = np.array(plot_position[3])

        numDataPoints = len(plot_position[0])
        dataSet = np.array([x, y, t])

        # NOTE: Can't pass empty arrays into 3d version of plot()
        line = plt.plot(dataSet[0], dataSet[1], dataSet[2], lw=2, c='g')[
            0]  # For line plot

        ax.set_xlabel('y')
        ax.set_ylabel('z')
        ax.set_zlabel('x')
        ax.set_title('Trajectory')

        # Creating the Animation object
        line_ani = animation.FuncAnimation(fig, func, frames=numDataPoints, fargs=(
            dataSet, line), interval=(time/numDataPoints), blit=False)
        plt.show()

    def show_all_graphs(self):
        """Displays all graphs."""
        plt.show()


class rocket_motor:

    """class representing a rocket motor"""

    def __init__(self, filePath: str, timeStep: int, ignitionTime: float = -1.0) -> None:
        """__init__ initializes the motor object

        Args:
            filePath (str, optional): the path to the motor file. Defaults to "".
            timeStep (int, optional): number of data points per second. Defaults to 0.
            ignitionTime (float, optional): ignition time of the motor. Defaults to -1.0.
        """
        self._ignitionTime: float = ignitionTime
        self._timeStep: int = timeStep
        self._data = []

        if filePath != "":

            tree = ET.parse(filePath)
            root = tree.getroot()

            eng_data = root[0][0][1]

            lPoint = [0, 0, 0]

            for data in eng_data:
                dataTmp = [float(data.attrib['t']), float(
                    data.attrib['f']), float(data.attrib['m'])]
                if dataTmp[0] > 0:
                    thrustDiff = dataTmp[1] - lPoint[1]
                    massDiff = dataTmp[2] - lPoint[2]
                    timeDiff = dataTmp[0] - lPoint[0]
                    stepsNeeded = timeDiff * timeStep

                    if stepsNeeded > 0:
                        adder = thrustDiff / stepsNeeded
                        adder_mass = massDiff / stepsNeeded

                        i = 0

                        thrustToAdd = lPoint[1]
                        massToAdd = lPoint[2]

                        while i < stepsNeeded:
                            i += 1
                            thrustToAdd += adder
                            if thrustToAdd < 0.0:
                                thrustToAdd = 0.0
                            massToAdd += adder_mass
                            if massToAdd < 0.0:
                                massToAdd = 0.0
                            self._data.append([thrustToAdd, massToAdd])
                lPoint = dataTmp

    def set_ignition_time(self, time: float) -> None:
        """set_ignition_time: sets the ignition time of the motor

        Args:
            time (float): the ignition time
        """
        self._ignitionTime = time

    def update(self, time: float) -> tuple[float, float]:
        """update: calculates and returns the motor's thrust and mass values based on the time passed in

        Args:
            time (float): current time

        Returns:
            tuple (float, float): thrust and mass values
        """
        if time > self._ignitionTime and self._ignitionTime != -1.0:
            idx = int(time*self._timeStep)
            if idx < len(self._data):
                return self._data[idx][0], self._data[idx][1]*0.001
            else:
                return 0.0, self._data[-1][1]*0.001
        else:
            return 0.0, self._data[0][1]*0.001
